#include "cppa/cppa.hpp"
#include <iostream>
#include <set>
#include "cppa/detail/to_uniform_name.hpp"

using namespace std;
using namespace cppa;

constexpr uint32_t sheep    = 0x0001;
constexpr uint32_t wolf     = 0x0002;
constexpr uint32_t cabbage  = 0x0004;
constexpr uint32_t shepherd = 0x0008;

typedef set<atom_value> atom_set;

void print(uint32_t state, bool print_endl = true) {
    if (state == 0) {
        aout << "- ";
        if (print_endl) aout << endl;
        return;
    }
    if (state & sheep) aout << "sheep ";
    if (state & wolf) aout << "wolf ";
    if (state & cabbage) aout << "cabbage ";
    if (state & shepherd) aout << "shepherd ";
    if (print_endl) aout << endl;
}

pair<uint32_t, uint32_t> move_left(uint32_t left, uint32_t right, uint32_t what) {
    what = what | shepherd;
    return {left | what, right & ~what};
}

pair<uint32_t, uint32_t> move_right(uint32_t left, uint32_t right, uint32_t what) {
    what = what | shepherd;
    return {left & ~what, right | what};
}

using move_fun = std::function<pair<uint32_t,uint32_t> (uint32_t, uint32_t, uint32_t)>;

void print(pair<uint32_t, uint32_t> p) {
    print(p.first, false);
    aout << "||| ";
    print(p.second, false);
    aout << endl;
}

bool invalid(uint32_t state) {
    if (state & shepherd) return false;
    return   ((state & sheep) && (state & wolf))
          || ((state & sheep) && (state & cabbage));
}

void master() {
    become(
        on_arg_match >> [](uint32_t lhs, uint32_t rhs) {
            if (rhs == 0) return atom("win");
            return invalid(lhs) || invalid(rhs) ? atom("fail") : atom("ok");
        }
    );
}

void solver(actor_ptr m, vector<pair<uint32_t,uint32_t>> visited, uint32_t lhs, uint32_t rhs) {
    self->link_to(m);
    vector<uint32_t> possible_moves;
    auto fun = (lhs & shepherd) ? move_right : move_left;
    auto pos = (lhs & shepherd) ? lhs : rhs;
    possible_moves.push_back(0);
    if (pos & sheep) possible_moves.push_back(sheep);
    if (pos & cabbage) possible_moves.push_back(cabbage);
    if (pos & wolf) possible_moves.push_back(wolf);
    for (auto& pm : possible_moves) {
        spawn([=] {
            auto mv = fun(lhs, rhs,  pm);
            auto i = find(begin(visited), end(visited), mv);
            if (i != end(visited)) return;
            sync_send(m, mv.first, mv.second).then(
                on(atom("win")) >> [=] {
                    for (auto s : visited) print(s);
                    print(mv);
                    aout << "win" << endl;
                    send_exit(m, exit_reason::user_shutdown);
                },
                on(atom("ok")) >> [=] {
                    auto v = visited;
                    v.push_back(mv);
                    spawn(solver, m, move(v), mv.first, mv.second);
                },
                on(atom("fail")) >> [] { }
            );
        });
    }
}

int main() {
    auto m = spawn(master);
    spawn(solver, m, vector<pair<uint32_t,uint32_t>>{}, uint32_t{0}, sheep | wolf | cabbage | shepherd);
    await_all_others_done();
    shutdown();
}

